// 1 	ELEMENT_NODE
// 2 	ATTRIBUTE_NODE
// 3 	TEXT_NODE
// 4 	CDATA_SECTION_NODE
// 5 	ENTITY_REFERENCE_NODE
// 6 	ENTITY_NODE
// 7 	PROCESSING_INSTRUCTION_NODE
// 8 	COMMENT_NODE
// 9 	DOCUMENT_NODE
// 10 	DOCUMENT_TYPE_NODE
// 11 	DOCUMENT_FRAGMENT_NODE
// 12 	NOTATION_NODE

import XMLDOM from 'xmldom-ts';
let DOMParser = XMLDOM.DOMParserImpl;
// let XMLSerializer = XMLDOM.XMLSerializerImpl;
// let XMLDocument = XMLDOM.DocumentImpl;
// import base from './base.js';

export default class XML {

    // convenience function; sets parameters for XML.toString
    static minify(node) {
        return XML.toString(node, { indentation: '', lineBreaks: '', })
    }

    // convenience function; sets parameters for XML.toString
    static prettify(node) {
        return XML.toString(node, { indentation: '\t', lineBreaks: '\n', })
    }

    // convenience function; sets parameters for XML.toString
    static highlight(node) {
        return XML.toString(node, { indentation: '\t', lineBreaks: '\n', highlight: true })
    }




    /**
     * XML escaping for a given string.
     * @param {string} string 
     */
    static escape(string) {
        // remove everything forbidden by XML 1.0 specifications, plus the unicode replacement character U+FFFD
        let invalidCharacters1 = /([^\x09\x0A\x0D\x20-\uD7FF\uE000-\uFFFC\u{10000}-\u{10FFFF}])/ug;
        // remove everything not suggested by XML 1.0 specifications
        let invalidCharacters2 = /([\x7F-\x84]|[\x86-\x9F]|[\uFDD0-\uFDEF]|[\u{1FFFE}-\u{1FFFF}]|[\u{2FFFE}-\u{2FFFF}]|[\u{3FFFE}-\u{3FFFF}]|[\u{4FFFE}-\u{4FFFF}]|[\u{5FFFE}-\u{5FFFF}]|[\u{6FFFE}-\u{6FFFF}]|[\u{7FFFE}-\u{7FFFF}]|[\u{8FFFE}-\u{8FFFF}]|[\u{9FFFE}-\u{9FFFF}]|[\u{AFFFE}-\u{AFFFF}]|[\u{BFFFE}-\u{BFFFF}]|[\u{CFFFE}-\u{CFFFF}]|[\u{DFFFE}-\u{DFFFF}]|[\u{EFFFE}-\u{EFFFF}]|[\u{FFFFE}-\u{FFFFF}]|[\u{10FFFE}-\u{10FFFF}].)/ug;
        return string
            .replace(/&/g, '&amp;')
            .replace(/</g, '&lt;')
            .replace(/>/g, '&gt;')
            .replace(/"/g, '&quot;')
            .replace(invalidCharacters1, '')
            .replace(invalidCharacters2, '');
    }




    /**
     * determine the attribute type from a given value
     * can detect dates and numbers.
     * 'null', 'undefined', 'NaN', 'true', 'false' are also marked.
     * @param {any} val the value from which the type shall be derived
     */
    static valueType(val) {
        if (['null', 'undefined', 'NaN', 'true', 'false'].includes(val)) return val;
        if (val * 1 == val) return 'number';
        let date = new Date(val);
        if (date.getFullYear() > 1900 && date.getFullYear() < 2100) return 'date';
        if (date.getFullYear() > 1000 && date.getFullYear() < 5000) return 'far-date';
        return '';
    }




    /**
     * converts an XML object to string.
     * recursively traversing the XML tree.
     * supports a lot of formatting options
     * @param {XML || string} node pass and XML object or a string (which is first parsed, then converted back to string)
     * @param {object} options see function header
     * @param {number} level used internally to set the indentation level 
     */
    static toString(node, options, level = 0) {
        // merge options-parameter with default options
        options = {
            ...{
                indentation: '\t',
                lineBreaks: '\n',
                highlight: false,
                escapeValues: true,
                skipEmptyNodes: true, /// invert this to emptyNodes: false
                skipEmptyAttributes: true,
                emptyNodes: false,
                emptyAttributes: false,
                inlineTags: ['b', 'i', 'u', 'q']
            }, ...options
        };
        // if(!node) return;
        if (typeof node == 'string') node = new DOMParser().parseFromString(node, 'text/xml');

        // set special characters according to options-parameter
        let tabs = Array(level + 1).fill('').join(options.indentation);
        let newLine = options.lineBreaks;
        let lt = options.highlight ? '&lt;' : '<';
        let gt = options.highlight ? '&gt;' : '>';

        // helper function to encapsulate tags
        let tag = (name, val, _class = '') => (options.highlight ? `<${name} ${_class ? `class='${_class}'` : ''}>${val}</${name}>` : val);

        // helper function to do XML escaping if setup in options
        let value = val => options.escapeValues ? this.escape(val) : val;

        // if node is text-node and content is only linebreaks or spaces, return empty string
        if (node.nodeType == 3 || node.nodeType == 4) // TEXT and CDATA nodes
            if (!node.textContent.trim()) return '';
            else return tabs + tag('content', value(node.textContent.trim())) + newLine;

        // only relevant for the uppermost DOM layer, which is not an xml-node
        if (!node.tagName) return Array.from(node.childNodes).map(child => XML.toString(child, options, level)).join('');

        // if node has no attributes and no content and emptyNodes-option is set to false => skip it
        if (!options.emptyNodes && (node.attributes.length == 0 || Array.from(node.attributes).filter(x => x.value).length == 0) && node.childNodes.length == 0) return '';

        // start output with tagname 
        let output = tabs + tag('control', lt) + tag('tag', node.tagName);

        // helper attribute, 
        let hasOnlyOneTextChild = ((node.childNodes.length == 1) && (node.childNodes[0].nodeType == 3) && !node.childNodes[0].textContent.trim().includes('\n'));
        let hasOnlyEmptyTextChild = hasOnlyOneTextChild && (node.childNodes[0].textContent.trim() == '');
        for (let i = 0; i < node.attributes.length; i++)
            if (!(options.skipEmptyAttributes && !node.attributes[i].value))
                output += ' ' + tag('pair', tag('attribute', node.attributes[i].name, node.attributes[i].name) + tag('control', '="') + tag('value', value(node.attributes[i].value), XML.valueType(node.attributes[i].value)) + tag('control', '"'));

        if ((node.childNodes.length == 0) || hasOnlyEmptyTextChild) return output + tag('control', ' /' + gt) + newLine;
        else output += tag('control', gt);
        if (!hasOnlyOneTextChild) output += newLine;
        for (let i = 0; i < node.childNodes.length; i++)
            if (hasOnlyOneTextChild) output += tag('content', value(node.childNodes[i].textContent.trim()), XML.valueType(node.childNodes[i].textContent.trim()));
            else output += XML.toString(node.childNodes[i], options, level + 1);
        return output + (hasOnlyOneTextChild ? '' : tabs) + tag('control', lt + '/') + tag('tag', node.tagName) + tag('control', gt) + newLine;
    }
}

